#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.AbsynUtils

open Absyn
open Util

let is_operator s = (Sugar.IsMangledOpName s) && (s <> "op_Equality")
  
let wrap_pf_typ' t = Typ_app (Const.pf_typ, withsort t Kind_unknown)

let wrap_pf_typ t = withsort (Typ_app (Const.pf_typ, t)) Kind_unknown

let strip_pf_typ t = match t.v with
  | Typ_app({v=Typ_const(tc);sort=_}, t') ->
      if Pretty.str_of_lident tc.v = Pretty.str_of_lident Const.pf_lid then t'
      else failwith (Printf.sprintf "strip_pf_t %s" (Pretty.strTyp t))
  | _ -> failwith (Printf.sprintf "strip_pf_t %s" (Pretty.strTyp t))

let Wt t = withsort t Kind_unknown
let W e = withsort e (Wt Typ_unknown)

let rctr = ref (int64 -1)
let Wmark e =
  let v = !rctr in 
    rctr := v - (int64 1);
    withinfo e (Wt Typ_unknown) v

let is_marked e = ((e.p < (int64 0)) && (e.p > (int64 -1500)))
let get_marker e = Printf.sprintf "%d" e.p

let bvar_to_exp bv = withsort (Exp_bvar bv) bv.sort
let bvd_to_exp bvd t = bvar_to_exp (bvd_to_bvar_s bvd t)

let hoist e (body:exp->exp) : exp' = 
  let id = genident (Some (e.p)) in
  let bvd = (id,id) in
    Exp_let(false, [(bvd, e.sort, e)],
            body (withinfo (Exp_bvar (bvd_to_bvar_s bvd e.sort)) e.sort e.p))

let mk_eqtest e1 e2 = 
  let eq = W (Exp_fvar (W Const.op_Eq)) in  
    W (Exp_app(W (Exp_app(eq, e1)), e2))
      
let proof_counter = ref 0
let proof_stats_file = new System.IO.StreamWriter ("queries/proofstats.txt")
let proof_stats_flag = ref false

let compute_proof_stats proofs = 
  let rec final_typ t = match t.v with
    | Typ_univ (_,_,t) 
    | Typ_fun (_,_,t) -> final_typ t
    | _ -> t in
  let compute_proof_stats proof = 
    let count_proof_rule i e = match e with
      | Exp_constr_app(v, _, _, _) -> 
          let t = v.sort in
          let ft = final_typ t in
          let i' = match ft.v with 
            | Typ_app({v=Typ_const v; sort=_; p=_}, _) when Sugar.lid_equals v.v Const.pf_lid -> i+1
          | _ -> i in
            (i', e)
      | _ -> (i, e) in
    let noop i t = (i,t) in
    let count, _ = descend_exp_map noop count_proof_rule 0 proof in
      proof_counter := !proof_counter + 1;
      Printf.fprintf proof_stats_file "Proof %d uses %d proof rules\n" !proof_counter count;
      flush proof_stats_file in
    List.iter compute_proof_stats proofs

let rec is_value e = match e.v with 
  | Exp_constant _
  | Exp_bvar _
  | Exp_fvar _ 
  | Exp_abs _ 
  | Exp_tabs _ -> true
  | Exp_constr_app (_, _, _, el) -> List.for_all is_value el
  | Exp_proj (e, _) -> is_value e
  | Exp_recd (_, _, _, fn_e_l) -> List.for_all (fun (_,e) -> is_value e) fn_e_l
  | _ -> false

let rec unrefine t = match t.v with 
  | Typ_refine(x, t, _) -> unrefine t
  | _ -> t


let rec get_tycon t = match t.v with 
  | Typ_const tc -> Some t.v
  | Typ_app(t, _)
  | Typ_dep(t, _)
  | Typ_refine(_, t, _) -> get_tycon t 
  | Typ_btvar _ -> Some t.v
  | _ -> None

let fold_app_tapp = 
  List.fold_left (fun e -> function 
                    | Inl t -> W (Exp_tapp(e, t))
                    | Inr e' -> W (Exp_app(e, e')))
