#light "off"
// (c) Microsoft Corporation. All rights reserved
module Microsoft.Fine.Anf

open Absyn
open AbsynUtils
type anf_env = {enclosing_let:option<bvdef>}
let empty_env = {enclosing_let=None}

let p_hoist env e k = match env.enclosing_let, k with 
  | None, None
  | Some _, None  -> e
  | Some _, Some k -> k e
  | None, Some k -> withpos (hoist e k) e.p
      
let rec anf_exp env e: option<(exp -> exp)> -> exp = 
  let anf_exp_l (el:list<exp>) (c:list<exp> -> exp) = 
    let cl = List.map (anf_exp empty_env) el in
    let rec builder names = function
      | [] -> c (List.rev names)
      | cx::rest -> cx (Some (fun x -> builder (x::names) rest)) in
      builder [] cl in
  let apply k e = match k with
      None -> e
    | Some k -> k e in
    match e.v with 
      | Exp_bvar _ 
      | Exp_fvar _
      | Exp_constant _ 
      | Exp_gvar _
      | Exp_bot -> (fun k -> apply k e)
      | Exp_constr_app(dc, tl, phantoms, el) -> 
          (fun k -> 
             (anf_exp_l el 
                (fun xl -> 
                   let app =  withpos (Exp_constr_app (dc, tl, phantoms, xl)) e.p in
                     apply k app)))
      | Exp_abs (bvd, t, body) -> 
          (fun k -> 
             let abs = withinfo (Exp_abs(bvd, t, anf_exp empty_env body None)) e.sort e.p in
               apply k abs)
      | Exp_tabs (bvd, kind, e) -> 
          (fun k -> 
             let tabs = withinfo (Exp_tabs(bvd, kind, anf_exp empty_env e None)) e.sort e.p in
               apply k tabs)
      | Exp_primop(id, args) -> 
          (fun k -> 
             anf_exp_l args 
               (fun xl -> 
                  let primop = (withpos (Exp_primop(id, xl)) e.p) in
                    p_hoist env primop k))
      | Exp_app(e1, e2) -> 
          (fun k -> 
             anf_exp empty_env e1 
               (Some (fun x1 -> 
                        anf_exp empty_env e2 
                          (Some (fun x2 -> 
                                   let app = withinfo (Exp_app (x1, x2)) e.sort e.p in
                                     p_hoist env app k)))))
      | Exp_tapp(e, t) -> 
          (fun k -> 
             (anf_exp empty_env e) 
               (Some (fun x -> 
                        let tapp = withinfo (Exp_tapp (x, t)) e.sort e.p in
                          p_hoist env tapp k)))
      | Exp_match(e, branches, def) -> 
          (fun k -> 
             (anf_exp empty_env e)
               (Some (fun x -> 
                        let branches' = List.map (fun (pat, e) -> pat, anf_exp empty_env e None) branches in
                        let match' = withpos (Exp_match(x, branches', anf_exp empty_env def None)) e.p in
                          p_hoist env match' k)))
      | Exp_cond (e1, e2, e3) -> 
          (fun k -> 
             (anf_exp empty_env e1) 
               (Some (fun x -> 
                        let cond = withpos (Exp_cond(x, anf_exp empty_env e2 None, anf_exp empty_env e3 None)) e.p in
                          p_hoist env cond k)))
      | Exp_recd (lidopt, tl, el, fn_e_l) -> 
          (fun k -> 
             let fn, el = List.unzip fn_e_l in
               apply k (anf_exp_l el (fun xl -> withpos (Exp_recd (lidopt, tl, el, (List.zip fn xl))) e.p)))
      | Exp_proj(e, fn) -> 
          (fun k -> 
             (anf_exp empty_env e) 
               (Some (fun x -> (* projections are values in TC but we still hoist them to ease derefinement *)
                        let proj = withinfo (Exp_proj(x, fn)) e.sort e.p in
                          p_hoist env proj k)))
      | Exp_ascribed(e, t, ev) -> 
          (fun k -> 
             (anf_exp empty_env e) 
               (Some (fun x -> 
                        let asc = withinfo (Exp_ascribed(e, t, ev)) e.sort e.p in
                          apply k asc)))
      | Exp_let(isrec, [bvd, t, e], body) when isrec=false-> 
          (fun k -> 
             let env' = {enclosing_let=Some bvd} in
             let anf_let = anf_exp env' e
               (Some (fun x -> 
                        withinfo (Exp_let(isrec, [bvd, t, x], anf_exp empty_env body None)) e.sort e.p)) in
               p_hoist env anf_let k)
      | Exp_let(isrec, bindings, body) when isrec=true-> 
          (fun k -> 
             let bindings' = List.map (fun (bvd, t, e) -> (bvd, t, anf_exp empty_env e None)) bindings in
             let elet = withinfo (Exp_let(isrec, bindings', anf_exp empty_env body None)) e.sort e.p in
               p_hoist env elet k)
      | Exp_extern_call (eref, ident, t, tl, el) -> 
          (fun k -> 
             (anf_exp_l el (fun xl -> 
                              let call = withpos (Exp_extern_call (eref, ident, t, tl, xl)) e.p in
                                p_hoist env call k)))

let anf_module m = 
  let lbs = List.map 
    (fun (bindings, isrec) -> 
       let bindings' = List.map (fun (bvd, t, e) -> (bvd, t, anf_exp empty_env e None)) bindings in
         (bindings', isrec)) m.letbindings in
  let main = Util.bind_opt m.main (fun main -> Some (anf_exp empty_env main None)) in
  let anf_m = {m with letbindings=lbs; main=main} in
(*   let _ = *)
(*     if not (Sugar.lid_equals m.name Const.prims_lid) then *)
(*       (Printf.printf "Desugared module before ANF:\n"; Pretty.printModule m; *)
(*        Printf.printf "\nANF module:\n"; Pretty.printModule anf_m); *)
(*   in *)
    anf_m
