module Microsoft.Fine.Const

open Sugar
open Absyn


let p2l l = path_to_lid dummyRange l

let prims_lid = p2l ["Prims"]
  
let mk_typ path k = 
  let tc = p2l path in
  let tycon = withsort (Typ_const (withsort tc k)) k in
    tc, tycon

let kind_un_op = Kind_tcon(Kind_star, Kind_star)
let kind_bin_op = Kind_tcon(Kind_star, Kind_tcon(Kind_star, Kind_star)) 

let extern_id = Sugar.extern_id
let bool_lid, bool_typ = mk_typ ["Prims"; "bool"] Kind_star
let exp_true_bool = withsort (Exp_constant (Sugar.Const_bool true)) bool_typ
let exp_false_bool = withsort (Exp_constant (Sugar.Const_bool false)) bool_typ
let unit_lid, unit_typ = mk_typ ["Prims"; "unit"] Kind_star
let string_lid, string_typ = mk_typ ["Prims"; "string"] Kind_star
let int_lid, int_typ = mk_typ ["Prims"; "int"] Kind_star
let true_lid, true_typ = mk_typ ["Prims"; "true_prop"] Kind_star
let false_lid, false_typ = mk_typ ["Prims"; "false_prop"] Kind_star
let eq_lid = p2l ["Prims"; "Eq"]

let reflexivity_lid = p2l ["Prims"; "Reflexivity"]
let symmetry_lid = p2l ["Prims"; "Symmetry"]
let transitivity_lid = p2l ["Prims"; "Transitivity"]
let eqiff_lid = p2l ["Prims"; "EqIff"]
let eqmono_lid = p2l ["Prims"; "EqMono"]
let monoprop_lid = p2l ["Prims"; "MonoProp"]
let tupproj1_lid = p2l ["Prims"; "TupleProj1"]
let tupproj2_lid = p2l ["Prims"; "TupleProj2"]
let tupeq1_lid = p2l ["Prims"; "TupleEq1"]
let tupeq2_lid = p2l ["Prims"; "TupleEq2"]
let tupmonoeq_lid = p2l ["Prims";  "TupleMonoEq"]
let ifftrans_lid = p2l ["ProofLib"; "_Trans"]
let nfa_exnot_lid = p2l ["Prims"; "NotAll_existsNot"]

(* types for logical connectives and proofs *)
let and_lid, and_typ = mk_typ ["Prims"; "l_and"] kind_bin_op
let or_lid, or_typ = mk_typ ["Prims"; "l_or"] kind_bin_op
let not_lid, not_typ = mk_typ ["Prims"; "l_not"] kind_un_op
let implies_lid, implies_typ = mk_typ ["Prims"; "l_implies"] kind_bin_op
let iff_lid, iff_typ = mk_typ ["Prims"; "l_iff"] kind_bin_op
let pf_lid, pf_typ = mk_typ ["Prims"; "pf"] (Kind_tcon (Kind_star, Kind_star))
  
let list_lid, list_typ = mk_typ ["List"; "list"] kind_un_op
  
(* let prims =  *)
(*   [bool_lid, bool_typ; *)
(*    unit_lid, unit_typ; *)
(*    string_lid, string_typ; *)
(*    int_lid, int_typ; *)
(*    true_lid, true_typ; *)
(*    false_lid, false_typ; *)
(*    and_lid, and_typ; *)
(*    or_lid, or_typ; *)
(*    implies_lid, implies_typ; *)
(*    not_licd, not_typ] *)

let w t = withsort t Kind_star

let op_Eq = p2l ["Prims"; "op_Equality"]

(* let prop_Eq = p2l ["Prims"; "prop_Equality"] *)
(* let prim_ops =  *)
(*   let name = genident None in *)
(*   let btvar0 = w (Typ_btvar (withsort (0, name, name) Kind_star)) in *)
(*     [op_Eq, w (Typ_univ((name,name), Kind_star,  *)
(*                         w (Typ_fun(None, btvar0,  *)
(*                                    w (Typ_fun(None, btvar0, bool_typ))))))] *)
      
let cons_lid = p2l ["List"; "Cons"]
let nil_lid = p2l ["List"; "Nil"]
let and_data_lid = p2l ["Prims"; "AndCons"]
let or_data_lid  = p2l ["Prims"; "OrCons"]
let not_data_lid = p2l ["Prims"; "NotCons"]
let tuple_UU_lid =  p2l ["Prims"; "Tuple_UU"]
let tuple_UA_lid =  p2l ["Prims"; "Tuple_UA"]
let tuple_AU_lid =  p2l ["Prims"; "Tuple_AU"]
let tuple_AA_lid =  p2l ["Prims"; "Tuple_AA"]
let dcil_tuple_UU_name = ["Prims"; "Tuple_UU"]
let dcil_tuple_UA_name = ["Prims"; "Tuple_UA"]
let dcil_tuple_AU_name = ["Prims"; "Tuple_AU"]
let dcil_tuple_AA_name = ["Prims"; "Tuple_AA"]
let dcil_tuple_UU_namestring = "Prims.Tuple_UU"
let dcil_tuple_UA_namestring = "Prims.Tuple_UA"
let dcil_tuple_AU_namestring = "Prims.Tuple_AU"
let dcil_tuple_AA_namestring = "Prims.Tuple_AA"
let name_proj_one = "One"
let tuple_proj_one = p2l ["Prims"; name_proj_one]
let name_proj_two = "Two"
let tuple_proj_two = p2l ["Prims"; name_proj_two]
(* Juan: hard-wired field names for projections of tuples!! *)
(* If the translation uses different strategies to generate target field names, *)
(* then the following two need to be changed *)
let dcil_proj_one_name = "Prims_One"
let dcil_proj_two_name = "Prims_Two"

let bind_pf_lid = p2l ["Prims"; "BindPf"]
let modus_ponens_lid = p2l ["Prims"; "ModusPonens"]
let and_elim_1_lid = p2l ["Prims"; "AndElim1"]
let and_elim_2_lid = p2l ["Prims"; "AndElim2"]
let and_intro_lid = p2l ["Prims"; "AndIntro"]

let implies_intro_lid = p2l ["Prims"; "ImpliesIntro"]
let or_intro1_lid = p2l ["Prims"; "OrIntro1"]

let op_And_lid = p2l ["Prims"; "_dummy_op_AmpAmp"]
let op_Or_lid = p2l ["Prims";"_dummy_op_BarBar"]
let op_Not_lid = p2l ["Prims";"_dummy_op_Negation"]

let op_Add_lid = p2l ["Prims";"_dummy_op_Addition"]
let op_Subtraction_lid = p2l ["Prims";"_dummy_op_Subtraction"]
let op_Multiply_lid = p2l ["Prims";"_dummy_op_Multiply"]
let op_GreaterThanOrEqual_lid = p2l ["Prims";"_dummy_op_GreaterThanOrEqual"]

let is_tuple_data_lid lid = match path_of_lid lid with 
  | ["Prims"; ntup] -> ntup.StartsWith("Tuple")
  | _ -> false

let tuple_data_lid k1 k2 = match unbox k1, unbox k2 with
  | Kind_star, Kind_star -> tuple_UU_lid
  | Kind_star, Kind_affine -> tuple_UA_lid
  | Kind_affine, Kind_star -> tuple_AU_lid
  | Kind_affine, Kind_affine -> tuple_AA_lid 
