module Microsoft.Fine.DesugarEnv

open Absyn

type binding = 
  | DSBinding_typ_var of ident * kind
  | DSBinding_var of ident * typ
  | DSBinding_tycon of ident * kind
  | DSBinding_pending_subst of ident * ident

type env 
(* = { *)
(*   curmodule: lident; *)
(*   modules:list<lident * modul>;  (\* already fully type checked modules *\) *)
(*   open_namespaces: list<lident>; (\* fully qualified names, in order of precedence *\) *)

(*   sigaccum:signature;            (\* type declarations being accumulated for the current module *\) *)
(*   localbindings:list<ident * binding>;  (\* local name bindings for name resolution *\) *)
(*   recbindings:list<binding>     (\* names bound by type t1 .. and t2 = .. *\) *)
(* } *)

(* type env *)

exception Not_found_binding of lident 
val set_extern_typ : env -> Sugar.externref -> env
val extern_typ : env -> bool
val extern_ref : env -> option<Sugar.externref>
val clear_extern_typ : env -> env
val env_as_modul : env -> option<lident> -> modul
val initial_env : list<lident*modul> -> lident -> env
val current_module: env -> lident
val push_binding_with_real_name : env -> binding -> ident -> env
val push_local_binding : env -> binding -> (env * ident)
val push_sigelt:env -> sigelt -> env
val push_rec_binding:env -> binding -> env
val push_namespace: env -> lident -> env
val generalize : env -> bool
val set_generalize : env -> bool -> env

val try_lookup_typ_var : env -> ident -> option<typ>
val try_lookup_typ_name : env -> lident -> option<typ>

val try_lookup_id : env -> ident -> option<exp>
val try_lookup_lid : env -> lident -> option<exp>
val try_lookup_datacon: env -> lident -> option<exp>
val try_lookup_field_name : env -> lident -> option<lident>
val try_lookup_val_decl_typ : env -> lident -> option<typ>
val unique_name : env -> lident -> bool
val unique_typ_name : env -> lident -> bool

val open_modules:env -> list<lident * modul>

val dump_env : env -> unit
