#light "off"
// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Expandtyp

open Absyn

let expand_tparams env tps = 
  List.map (fun tp -> match tp with 
              | Tparam_typ _ -> tp
              | Tparam_term (bvd, t) -> 
                  let t = Tcenv.expand_typ_rec env t in
                    Tparam_term (bvd, t)) tps

let expand_typ env t = Tcenv.expand_typ_rec env t

let expand_exp env e = Tcenv.expand_exp_rec env e
  
let rec expand_kind env k = match k with
  | Kind_star
  | Kind_affine -> k
  | Kind_tcon(k, k') -> Kind_tcon(k, expand_kind env k')
  | Kind_dcon(t, k) -> Kind_dcon(expand_typ env t, expand_kind env k)
  | Kind_qkind(qs, k) -> 
      let env = List.fold_left (fun env (bvd, k) -> 
                                  Tcenv.push_local_binding env (Tcenv.Binding_typ ((snd bvd), k))) env qs in
        Kind_qkind(qs, expand_kind env k)
  | _ -> raise Util.Impos

let rec expand_sigelt env = function
  | Sig_typ_abbrev _ -> 
      None

  | Sig_record_typ (lid, tps, kk, t) -> 
      Some(Sig_record_typ(lid, expand_tparams env tps, expand_kind env kk, expand_typ env t))

  | Sig_tycon_kind (lid, tps, k, b) -> 
      Some (Sig_tycon_kind(lid, expand_tparams env tps, expand_kind env k, b))
        
  | Sig_datacon_typ (lid, tps, t, b, aq, tcn) -> 
      Some (Sig_datacon_typ(lid, expand_tparams env tps, expand_typ env t, b, aq, tcn))

  | Sig_value_decl (lid, t) -> 
      Some (Sig_value_decl(lid, expand_typ env t))

  | Sig_extern_value (lang, lid, t) ->
      Some (Sig_extern_value (lang, lid, expand_typ env t))

  | Sig_extern_typ(eref, s) -> 
      match expand_sigelt env s with 
        | None -> Some (Sig_extern_typ (eref, s))
        | Some s -> Some (Sig_extern_typ (eref, s))
 
let expand_let_binding env lb =
  List.map (fun (bvd, t, e) -> 
              let t = expand_typ env t in
              let e = expand_exp env e in
                (bvd, t, e)) lb
    
let expand_typ_abbrevs env modul = 
  let env = Tcenv.clear_expand_externs env in
  let sigs = List.fold_right 
    (fun sigelt sigout -> 
       match expand_sigelt env sigelt with
           None -> sigout
         | Some se -> se::sigout) modul.signature [] in
  let lbs = List.map 
    (fun (lb, b) -> 
       (expand_let_binding env lb, b)) modul.letbindings in
  let main = match modul.main with 
      None -> None
    | Some e -> Some (expand_exp env e) in
    {modul with signature=sigs; letbindings=lbs; main=main}
