#light "off"
// (c) Microsoft Corporation. All rights reserved

  (* Known issues:

     1. Local vars don't work in static Initializers for non-function fields
       e.g., let initial_state = let a = Foo in a

     2. Unit is not translated properly. At the moment, it's
        translated to bool
 *)
module Microsoft.Fine.Ilgen

open Util
open Target
open Microsoft.FSharp.Collections
open Microsoft.FSharp.Compiler.AbstractIL
open Microsoft.FSharp.Compiler.AbstractIL.IL

let DEFAULT_MAX_STACK = 256

let unicodeEncoding = new System.Text.UnicodeEncoding()
type env = 
    { curmodulename : string;
      curclassname : string;
      entry:option<tExp>;
      classDeclEnv: TargetChecker.tCenv;
      externs: TargetChecker.tCenv;
      tvarindices: list<tIdent * uint16>;
      paramindices: list<tIdent * uint16>;
      localindices: list<tIdent * uint16>;
      fieldspecs: list<tIdent * (bool * ILFieldSpec)>;
      topfields:option<(ILTypeRef * ILType * list<tStaticFieldDecl>)>;
      eraseUnit:bool
    }

let strEnv env = "TODO"
  (*  spr "[%s]" (String.concat " ### " (List.map (fun cd -> Sugar.text_of_path cd.name) env.classDeclEnv))*)

let fsharp_core_1_9_6_2 = 
  ILAssemblyRef.Create("FSharp.Core", 
                       None, 
                       Some (PublicKeyToken [| 0xA1uy; 0x90uy; 0x89uy; 0xB1uy; 0xC7uy; 0x4Duy; 0x08uy; 0x09uy|]),
                       false, 
                       (Some (1us, 9us, 6us, 2us)),
                       None)
let fsharp_core_2_0_0_0 =
  ILAssemblyRef.Create("FSharp.Core",
                       None,
                       Some (PublicKeyToken [| 0xB0uy; 0x3Fuy; 0x5Fuy; 0x7Fuy; 0x11uy; 0xD5uy; 0x0Auy; 0x3Auy |]),
                       false,
                       (Some (2us, 0us, 0us, 0us)),
                       None)

let fsharp_core = fsharp_core_2_0_0_0 

exception Ilgen_error of string
let must o msg = match o with 
  | Some x -> x
  | None -> raise (Ilgen_error (msg()))

let or_else o f g = match o with
  | Some x -> f x
  | None -> g ()

let my_mk_simple_assref asname = 
  match asname with 
    | "FSharp.Core" -> fsharp_core
    | _ -> mk_simple_assref asname
      
let catch_not_found f = 
  try 
    Some (f ())
  with
    | :? System.Collections.Generic.KeyNotFoundException -> None 

let lookup_cdecl env cn : option<tClassDecl> =
  match cn with 
    | "__Extern"::rest -> 
        TargetChecker.tryLookupCdeclByName env.externs rest
    | _ -> 
        TargetChecker.tryLookupCdeclByName env.classDeclEnv cn

let lookup_tvar_id env tv =
  catch_not_found (fun _ -> List.assoc (fst tv) env.tvarindices)

let lookup_local_id env (tv:tVar<tType>) = 
  catch_not_found (fun _ -> List.assoc (fst tv) env.localindices)

let mk_index_map init (al:list<'a>) (f:'a -> 'b) : list<'b * uint16> = 
  let _, ixmap = List.fold_left (fun (ix, out) a -> (ix + 1us, (f a, ix)::out)) (init, []) al in  
    ixmap

let initial_env (modul:tModule) = 
  let mname = Sugar.text_of_path modul.name in
  let cenv = [modul.decls;Target.predefined] in
  let e = modul.entry in
    {curmodulename=mname; classDeclEnv=cenv; externs=[modul.externs]; tvarindices=[]; 
     paramindices=[]; localindices=[]; fieldspecs=[]; entry=e;
     topfields=None; curclassname="<NONE>"; eraseUnit=false
    }  

let add_generic_params env gpds = 
  let init = match env.tvarindices with 
      (_, init)::_ -> init + 1us
    | _ -> 0us in
  let ixmap = mk_index_map init gpds (fun gpd -> gpd.gpName) in
    {env with tvarindices=ixmap@env.tvarindices}

let add_parameter_map env plist = 
  let ixmap = mk_index_map 0us plist fst in
    {env with paramindices=ixmap}
      
let add_local_map env locals = 
  let ixmap = mk_index_map 0us locals fst in
    {env with localindices=ixmap}

let add_fieldspec_map env fsm = {env with fieldspecs=fsm}

let extend_env env (modul:tModule) = 
  let mname = Sugar.text_of_path modul.name in
  let cenv = modul.decls in
  let e = modul.entry in
    {curmodulename=mname; classDeclEnv=cenv::env.classDeclEnv; externs=[modul.externs]; curclassname="<NONE>";
     tvarindices=[]; paramindices=[]; localindices=[]; fieldspecs=[]; entry=e; topfields=None; eraseUnit=false}  


let lookup_failed msg a = 
  Printf.printf "Lookup failed %s\n" msg; printfn "%A" a;
  raise (Ilgen_error msg)
    
let lookup_local env tv =
    catch_not_found (fun _ -> List.assoc (fst tv) env.localindices)
    
(* 0th param is the this ptr, unless this is a static method *)        
let lookup_param env tv = 
  catch_not_found (fun _ ->
                     let x = List.assoc (fst tv) env.paramindices in
                       x + 1us)     
       
let is_extern_name (cname:tClassName) = 
  match cname with
    | "__Extern"::dll::rest -> true 
    | _ -> false

let get_extern_name (cname:tClassName) = 
  match cname with
    | "__Extern"::dll::rest -> Some (dll,rest)
    | _ -> None

let lookup_field env tconc fref = 
  catch_not_found 
      (fun _ -> 
         let cname, ts, vs = tconc in
           match is_extern_name cname with 
             | false ->
                 let cdecl = must (lookup_cdecl env cname) (fun () -> "Class not found") in
                 let fname, ftype = List.find (fun (fn, t) -> fn=fref) (cdecl.vvars@cdecl.fields) in
                 let out, _ = List.fold_left (fun (out,ix) tv -> TType_index ix::out, ix+1us) ([],0us) (cdecl.ptvars@cdecl.tvars) in
                 let tsub = List.rev out in
                 let ftypeInst = TargetChecker.instantiateType cdecl tsub vs ftype in
                   fname, ftypeInst
             | true -> 
                 let cdecl = must (lookup_cdecl env cname) (fun () -> "Class not found") in
                 let fname, ftype = List.find (fun (fn, t) -> fn=fref) cdecl.vvars in
                   fname, ftype)

let lookup_method env tconc mref = 
  catch_not_found (fun _ -> 
  let cname, ts, vs = tconc in
  let cdecl = TargetChecker.lookupCdeclByName env.classDeclEnv cname in
  let (retType, mname, targs, params, body) = 
    List.find (fun (_, name, _, _, _) -> (name = fst mref)) cdecl.methods in
  let targs = cdecl.ptvars@cdecl.tvars@targs in
  let tsubst = List.rev (snd (List.fold_left (fun (ix, out) t -> ix+1us, TargetChecker.TvarMap(t, TType_index ix)::out) (0us, []) targs)) in
  let tmap = tsubst in (* (TargetChecker.buildTyMap cdecl ts vs)@tsubst in *)
  let retType = TargetChecker.substType tmap retType in
  let params = List.map (fun (n,t) -> n, TargetChecker.substType tmap t) params in
    (retType, mname, targs, params, None))
          
type fieldqual = Instance | Static

let mscorlibrefs:ILGlobals = mk_ILGlobals ecma_mscorlib_scoref None
  
let props = IL.mk_properties []
let atts = IL.mk_custom_attrs []
let events = IL.mk_events []

let is_value_type = function
  | ["Prims"; "int"]
  | ["Prims"; "bool"] -> true
  | _ -> false

let tClassName2NamespaceAndString = Util.pfx 

let tVisibility2TypeDefAccess = function
  | TVis_public -> IL.Public
(*   | TVis_internal -> TypeAccess_private *)
  | TVis_internal -> IL.Nested (MemAccess_family)

let tVar2GenericParameterDef (tvarName, tvarKind) = 
  (* does not always have to be ref type; value types like int etc. are permissible too *)
  (* ignoring kind constraints *)
  {gpName=tvarName; 
   gpCustomAttrs=mk_custom_attrs [];
   gpConstraints=[];  (* TODO: add custom modifier here for tvarKind *)
   gpVariance=NonVariant;
   gpReferenceTypeConstraint=false; 
   gpNotNullableValueTypeConstraint=false;
   gpDefaultConstructorConstraint=false}

let is_primitive_deparrow = function 
    ["DepArrowSS"]
  | ["DepArrowSA"]
  | ["DepArrowAS"]
  | ["DepArrowAA"] -> true
  | _ -> false


let is_tuple_class c =
  c=Const.dcil_tuple_UU_name ||
  c=Const.dcil_tuple_UA_name ||
  c=Const.dcil_tuple_AU_name ||
  c=Const.dcil_tuple_AA_name ||
  c=Const.dcil_tuple_UU_name ||
  c=Target.tClassDepTuple2SS.name ||
  c=Target.tClassDepTuple2SA.name ||
  c=Target.tClassTuple2AS.name ||
  c=Target.tClassTuple2AA.name 
  
let is_primitive = function 
    "DepArrowSS"
  | "DepArrowSA"
  | "DepArrowAS"
  | "DepArrowAA"
  | "Affine"
  | "DepTuple2SS"
  | "DepTuple2SA"
  | "Tuple2SS"
  | "Tuple2SA"
  | "Tuple2AA"
  | "Tuple2AS" 
  | "All**"
  | "All*A"
  | "AllA*"
  | "AllAA" -> true
  | _ -> false
      
let tryfind_extern_classdecl env = function
  | "__Extern"::fullname ->  TargetChecker.tryLookupCdeclByName env.externs fullname
  | _ -> None

let strenv [e] =
  let keys = Hashtbl.fold (fun key value st -> key::st) e [] in
    String.concat ", " keys 

let tClassName2TypeRef env tlid ngenerics = 
  let map_extern_class_name sr ns cn = 
    let fullname = ns@[cn] in
    let enclosingClass = List.tl ns in
    let tref = mk_tref(sr, Sugar.text_of_path enclosingClass) in
      match TargetChecker.tryLookupCdeclByName env.externs (["__Extern"]@fullname) with 
        | None -> pr "Unable to find class %s in environment [%s]\n" (PrettyTarget.strLident fullname) (strenv env.externs); raise Impos
        | Some cdecl -> 
            match cdecl.extends with
                None -> 
                  mk_tref_in_tref(tref, cn)
              | Some (TType_name((tcn,_))) -> (* we have a constructor class and need to mangle names *)
                  let _, parent = Util.pfx tcn in
                  let tref = mk_tref_in_tref (tref, parent) in
                  let cn = "_"^cn in
                    mk_tref_in_tref(tref, cn) in
  let ns_cn tlid = 
    let ns, cn = Util.pfx tlid in
    let ns =
      if ns=[] then
        if is_primitive cn then ["Prims"] else [env.curmodulename]
      else ns in 
      ns, cn in
    (* TODO: Special handling for primitive types; map these to mscorlib types *)
  let ns, cn = ns_cn tlid in
    match ns, cn with 
      | ["Prims"], "string" -> mscorlibrefs.tref_String
      | ["Prims"], "int" -> mscorlibrefs.tspec_Int32.TypeRef 
      | ["Prims"], "bool" -> mscorlibrefs.tspec_Bool.TypeRef 
      | ["Prims"], "unit" -> mscorlibrefs.tspec_Bool.TypeRef (* units are bools for now *)
      | ["Prims"], "object"
      | ["System"], "object" 
      | ["Prims"], "Affine" -> mscorlibrefs.tref_Object
      | "__Extern"::dll::[classname], innerclass -> 
          let assref = my_mk_simple_assref dll in
          let sr = ScopeRef_assembly assref in
            map_extern_class_name sr (dll::[classname]) innerclass 
      | ["Prims"], "DepArrowSS" -> 
          let dll = "FSharp.Core" in 
          let assref = my_mk_simple_assref dll in
          let sr = ScopeRef_assembly assref in
            mk_nested_tref(sr, [], "Microsoft.FSharp.Core.FSharpFunc`2")
      | ["Microsoft";"FSharp";"Core"], _ ->
          let dll = "FSharp.Core" in 
          let assref = my_mk_simple_assref dll in
          let sr = ScopeRef_assembly assref in
            mk_nested_tref(sr, [], String.concat "." (ns@[cn]))
      | _ -> 
          let assemblyName = Sugar.text_of_path ns in
          let cn = if ngenerics <> 0 then spr "%s`%d" cn ngenerics else cn in
(*             Printf.printf "Making reference to class %s\n" cn; *)
            if assemblyName = env.curmodulename then 
              ILTypeRef.Create (ScopeRef_local, [env.curmodulename], cn)
(*               TypeRef.Create (ScopeRef_local, [], cn)  *)
            else
              let _ = trace -1831 "tcn2typeref for " (tlid, assemblyName) in
              let assRef = my_mk_simple_assref assemblyName in
              let sr = ScopeRef_assembly assRef in
              let fcn = cn in (* classes are nested inside a module main class *)
                if cn = assemblyName then 
                  mk_tref(sr, fcn) 
                else
                  mk_nested_tref(sr, [assemblyName], fcn)
(*                 mk_tref(sr, fcn) (\* Q: Should cn be the full name or just the class name? *\) *)
                  
let typeName2TypeSig (tn, _) = 
  match Util.pfx tn with 
      | ["Prims"], "String" -> Some mscorlibrefs.typ_String
      | ["Prims"], "int" -> Some mscorlibrefs.typ_int32
      | ["Prims"], "bool" -> Some mscorlibrefs.typ_bool
      | ["Prims"], "unit" -> (* Some Type_void *)Some mscorlibrefs.typ_bool (* units are bools for now *)
      | [], "Affine" -> Some mscorlibrefs.typ_Object
      | _ -> None

let first_order_tvars = 
  List.filter (fun (_,k) -> match k with (* remove higher-order tvars *)
                 | TKind_karrow _ -> false
                 | _ -> true) 

let rec tTypeConcrete2TypeSpec env (tn, targs, _) = (* Note: value args are erased *)
  let gargs = 
    List.fold_left (fun out t -> match tType2TypeSigOpt env t with 
                      | None -> out
                      | Some t' -> t'::out) [] targs in
  let ngenerics = List.length gargs in
  let tr = tClassName2TypeRef env tn ngenerics in
    ILTypeSpec.Create(tr, List.rev gargs)
      
and lookup_fieldspec env tv = 
  catch_not_found (fun _ -> 
                     match List.try_assoc (fst tv) env.fieldspecs with 
                       | Some x -> x
                       | None -> (* try looking this up as a top-level binder *)
                           match env.topfields with
                             | None -> raise Not_found
                             | Some (tref, tsig, fields) -> 
                                 match List.tryFind (fun (fn, t, _) -> fn=(fst tv)) fields with
                                   | None -> 
                                       pr "Looking for field named %s in class %s .. not found\n"
                                         (fst tv) (env.curmodulename);
                                       raise Not_found
                                   | Some (fn,t,_) -> 
                                       let fieldtsig = tType2TypeSig env t in
                                       let fref = mk_fref_in_tref(tref, fn, fieldtsig) in
                                         true, mk_fspec(fref, tsig))
    
and tType2TypeSigOpt env = function
  | TType_index i -> Some (Type_tyvar i)
      
  | TType_var tv -> 
      let id = must (lookup_tvar_id env tv) (fun () ->  (spr "Type variable %s not found when checking class %s\n" (PrettyTarget.strTvar tv) env.curclassname)) in
        Some (Type_tyvar id)

  | TType_name tn -> 
      (match typeName2TypeSig tn with 
         | None -> 
             let ts = tTypeConcrete2TypeSpec env (fst tn, [], []) in
               if is_value_type (fst tn) then Some (Type_value ts)
               else Some (Type_boxed ts)
         | Some ts -> Some ts)
                 
  | TType_fun (bvd, t1, t2) -> (* Type-level functions \x:t1.t2 are erased to the erasure of t2 *)
      tType2TypeSigOpt env t2 

  | TType_dep (t, _) -> (* Dependent types (t v) are erased to the erasure of t *)
      tType2TypeSigOpt env t 
        
  | TType_concrete ct -> 
      let ts = tTypeConcrete2TypeSpec env ct in
        Some (Type_boxed ts)
      
  | TType_affine t -> (* affine types !t are erased to the erasure of t *)
      Some (tType2TypeSig env t)

  | TType_tfun _ -> 
      (* these are expected to occur only among the targs of a concrete type 
         In particular, in the arguments of Id : All<\c. Fun<'c,'c>> etc. 
         These are erased altogether *)
      None
        
  | TType_tapp _ -> 
      (* these are expected to occur only in the return type of TyApp in a sub-class of All 
         And are erased to object *)
      Some mscorlibrefs.typ_Object

  | t -> printfn "%A" t; raise (Ilgen_error "Unexpected target type in ilgen")

and tType2TypeSig env t = 
  must (tType2TypeSigOpt env t) 
    (fun () -> (spr " match tType2TypSigOpt expected a translatable type; got %s\n" (PrettyTarget.strType t)))
    
let tFieldDecl2FieldDef fq env (fname, ftyp) = 
  let fTypeSig = tType2TypeSig env ftyp in
    match fq with 
      | Instance -> mk_instance_fdef (fname, fTypeSig, None, MemAccess_public)
      | Static -> mk_static_fdef (fname, fTypeSig, None, None, MemAccess_public)

let rec tTypeOftValue env = function
  | TVal_var (_, tt) -> tt
  | TVal_obj (ttc, _,  _) -> TType_concrete ttc
  | TVal_constant sc -> TargetChecker.constTyping sc
  | TVal_ldfld (tv, fr) -> let _, t = lookup_field_typ env (tTypeOftValue env tv) fr in t

and lookup_field_typ env vtype fref = 
  let msg i () = spr "[%d](In class %s) Field ref %s not found in type %s\n" i (env.curclassname) fref (PrettyTarget.strType vtype)in
  let fail i = raise (Ilgen_error (msg i ())) in
    match vtype with
      | TType_concrete tconc ->
          let (cname, ts, vs) = tconc in
            if (is_tuple_class cname) then 
              match ts with
                | [t1; TType_fun(_, _, t2)] -> 
                    if (fref = Const.dcil_proj_one_name) then 
                      let cd = 
                        must (lookup_cdecl env cname) 
                          (fun () -> (spr "Could not find class declaration for %s\n" (PrettyTarget.strLident cname))) in 
                      let (f_fst, _)::tl = cd.vvars in
                        must (lookup_field env tconc f_fst) (msg 1) 
                    else if (fref = Const.dcil_proj_two_name) then 
                      let cd = must (lookup_cdecl env cname) 
                        (fun () -> (spr "Could not find class declaration for %s\n" (PrettyTarget.strLident cname))) in 
                      let [_;(f_snd, _)] = cd.vvars in
                        must (lookup_field env tconc f_snd) (msg 1) 
                    else must (lookup_field env tconc fref) (msg 1) 
                | _ -> must (lookup_field env tconc fref) (msg 2) 
            else  must (lookup_field env tconc fref) (msg 3)
      | TType_name (n,_) when not (is_extern_name n) -> 
          must (lookup_field env (n,[],[]) fref) (msg 4)
      | TType_name (n,_) -> 
          (match get_extern_name n with 
             | None -> raise Impos
             | Some (dll, classname) -> 
                 let cdecl = TargetChecker.lookupCdeclByName  env.externs (dll::classname) in
                   must (List.tryFind (fun (fn, t) -> fn=fref) (cdecl.fields)) (msg 5))
      | _ -> fail 6
          
and tExp2Code env exp = 
  let sconst2Instr = function 
    | Sugar.Const_unit (* Translate this as a boolean for now. TODO: Erase booleans to void. *)
    | Sugar.Const_bool false -> 
        I_arith (AI_ldc (DT_I4, NUM_I4 0))

    | Sugar.Const_bool true -> 
        I_arith (AI_ldc (DT_I4, NUM_I4 1))

    | Sugar.Const_int8 i -> 
        I_arith (AI_ldc (DT_I1, NUM_I4 (int32 i)))
        
    | Sugar.Const_uint8 i -> 
        I_arith (AI_ldc (DT_U1, NUM_I4 (int32 i)))

    | Sugar.Const_int16 i -> 
        I_arith (AI_ldc (DT_I2, NUM_I4 (int32 i)))

    | Sugar.Const_uint16 i -> 
        I_arith (AI_ldc (DT_U2, NUM_I4 (int32 i)))

    | Sugar.Const_int32 i -> 
        I_arith (AI_ldc (DT_I4, NUM_I4 (int32 i)))
          
    | Sugar.Const_uint32 i -> 
        I_arith (AI_ldc (DT_U4, NUM_I4 (int32 i)))

    | Sugar.Const_nativeint i 
    | Sugar.Const_int64 i -> 
        I_arith (AI_ldc (DT_I8, NUM_I8 i))

    | Sugar.Const_unativeint i
    | Sugar.Const_uint64 i -> 
        I_arith (AI_ldc (DT_U8, NUM_I8 (int64 i)))

    | Sugar.Const_float32 i -> 
        I_arith (AI_ldc (DT_R4, NUM_R4 i))

    | Sugar.Const_float i -> 
        I_arith (AI_ldc (DT_R8, NUM_R8 i))

    | Sugar.Const_char i -> 
        I_arith (AI_ldc (DT_I1, NUM_I4 (int32 i)))

    | Sugar.Const_string (bytes, _) -> 
        let s = unicodeEncoding.GetString(bytes) in
        I_ldstr (s)
        
    | Sugar.Const_decimal _
    | Sugar.Const_bigint _ 
    | Sugar.Const_bignum _
    | Sugar.Const_bytearray _ 
    | Sugar.Const_uint16array _ -> raise (NYI "Unhandled constant form") in
        
  let rec tExp2Instrs : tExp -> list<ILInstr> = function
    | TExp_val (TVal_constant sc) ->
        (match sc with
             Sugar.Const_unit when env.eraseUnit -> [] (* HACK! *)
           | _ -> [sconst2Instr sc])
          
    | TExp_val (TVal_var tv) -> 
        or_else 
          (lookup_local env tv) 
          (fun local_ix -> [I_ldloc local_ix]) 
          (fun _ -> 
             or_else 
               (lookup_param env tv)
               (fun param_ix -> [I_ldarg param_ix])
               (fun _ -> 
                  let isStatic, fs = must (lookup_fieldspec env tv) (fun () -> (spr "Unable to translate variable: %s" (fst tv))) in
                    if isStatic then 
                      [mk_normal_ldsfld fs]
                    else 
                      [I_ldarg 0us; (* load this *)
                       mk_normal_ldfld fs]))
          
    | TExp_name (tn, tt) ->
        let assemblyName, fieldName = Util.pfx tn in 
        let assemblyNameStr = String.concat "." assemblyName in
        let assref = ScopeRef_assembly(my_mk_simple_assref assemblyNameStr) in
          (*         let containerClass = assemblyName in (\* @ ["_Container"] in *\) *)
        let containerTref = mk_tref(assref, assemblyNameStr) in 
        let fieldType = tType2TypeSig env tt in
        let fspec = mk_fspec_in_nongeneric_boxed_tref(containerTref, fieldName, fieldType) in
          [mk_normal_ldsfld fspec]

    | TExp_val(TVal_ldfld(v, fref)) 
    | TExp_ldfld (v, fref) -> 
(*         let msg () = spr "Field ref %s not found" fref in  *)
        let instrs = tExp2Instrs (TExp_val v) in
        let vtype = tTypeOftValue env v in
        let vtypeSig = tType2TypeSig env vtype in
        let fname, ftype = lookup_field_typ env vtype fref in
        let ld_or_call = match vtype with
          | TType_name (n,_) when is_extern_name n-> 
              (match get_extern_name n with 
                 | None -> raise Impos
                 | Some (dll, classname) -> 
                     let ftsig = tType2TypeSig env ftype in
                     let mname = "get_"^fname in
                     let cc = Callconv(CC_instance, CC_default) in
                     let methodRef = mk_mref(tref_of_typ vtypeSig, cc, mname, 0, [], ftsig) in
                     let mspec = mk_mref_mspec_in_typ(methodRef, vtypeSig, []) in
                       mk_normal_callvirt mspec)
          | _ -> 
              let ftypeSig = tType2TypeSig env ftype in 
              let fspec = mk_fspec_in_typ (vtypeSig, fname, ftypeSig) in
                mk_normal_ldfld fspec
        in                  
          instrs@[I_castclass vtypeSig; ld_or_call]

(*     | TExp_extern_call (eref, (mn, []), argTypes, retType, args) ->  *)
(*         let assref = ScopeRef_assembly(mk_simple_assref eref.dll) in *)
(*         let tref = mk_tref (assref, Sugar.text_of_lid eref.classname) in *)
(*         let tsig = mk_boxed_typ tref [] in *)
(*         let argTsigs = List.map (tType2TypeSig env) argTypes in *)
(*         let retTsig = tType2TypeSig env retType in *)
(*         let mspec = mk_static_nongeneric_mspec_in_typ (tsig, mn, argTsigs, retTsig) in *)
(*         let callInstr = mk_normal_call mspec in *)
(*         let argLdInstrs = List.concat (List.map (fun argv -> tExp2Instrs (TExp_val argv)) args) in *)
(*           argLdInstrs@[callInstr] *)

    | TExp_extern_call (eref, (mn, tl), argTypes, retType, args) -> 
        let assref = ScopeRef_assembly(my_mk_simple_assref eref.dll) in
        let tref = mk_tref (assref, Sugar.text_of_lid eref.classname) in
        let tref = match eref.innerclass with 
            None -> tref
          | Some s -> mk_tref_in_tref (tref, s) in
        let tsig = mk_boxed_typ tref [] in
        let argTsigs = List.map (tType2TypeSig env) argTypes in
        let retTsig = tType2TypeSig env retType in
        let genArgs = List.map (tType2TypeSig env) tl in
        let mspec = mk_static_mspec_in_typ (tsig, mn, argTsigs, retTsig, genArgs) in
        let callInstr = mk_normal_call mspec in
        let argLdInstrs = List.concat (List.map (fun argv -> tExp2Instrs (TExp_val argv)) args) in
          argLdInstrs@[callInstr]

    | TExp_primop(op, args) ->
        let argLdInstrs = List.concat (List.map (fun arg -> tExp2Instrs (TExp_val arg)) args) in
        let op_instr = match op with 
          | AND -> AI_and
          | OR -> AI_or
          | ADD -> AI_add
          | SUB -> AI_sub
          | MUL -> AI_mul
          | DIV -> AI_div
          | NOT -> AI_not in
          argLdInstrs@[I_arith op_instr]

    | TExp_call (v, [arg], ("Equals", [t])) when t=Target.objType -> 
        let instrs = tExp2Instrs (TExp_val v) in
        let argLdInstrs = tExp2Instrs (TExp_val arg) in
        let instrs = instrs@argLdInstrs in
        let vtype = tTypeOftValue env v in
        let vtypeSig = tType2TypeSig env vtype in
        let vtypeRef = tref_of_typ vtypeSig in
        let genArity = 0 in 
        let retTypeSig = mscorlibrefs.typ_Bool in
        let plist = [mscorlibrefs.typ_Object] in
        let cc = Callconv(CC_instance, CC_default) in
        let mspec = mk_nongeneric_mspec_in_typ(vtypeSig, cc, "Equals", plist, retTypeSig) in
        let callInstr = mk_normal_callvirt mspec in
          instrs@[callInstr]

    | TExp_call(v, args, ((mname, tparams) as tmref)) when mname="Invoke" -> 
        let instrs = tExp2Instrs (TExp_val v) in
        let argLdInstrs = List.concat (List.map (fun argv -> tExp2Instrs (TExp_val argv)) args) in
        let instrs = instrs@argLdInstrs in
        let vtype = tTypeOftValue env v in
        let (retType, mname, targs, params, _) = match vtype with
          | TType_affine (TType_concrete((tcn, _, _) as tconc))
          | TType_concrete ((tcn, _, _) as tconc) when is_primitive_deparrow tcn ->
              must (lookup_method env tconc tmref) (fun () -> (spr "Method ref %s not found in class %s\n" mname (String.concat "." tcn)))
          | _ -> raise Impos in
        let vtypeSig = tType2TypeSig env vtype in
        let vtypeRef = tref_of_typ vtypeSig in
        let genArity = List.length targs in
        let retTypeSig = (Type_tyvar 1us) in (* notice that numbering of return type ... *)
        let plist = [Type_tyvar 0us] in (* ... and formal is relative to the static decl of FastFunc. *)
        let cc = Callconv(CC_instance, CC_default) in
        let methodRef = mk_mref(vtypeRef, cc, mname, genArity, plist, retTypeSig) in
        let genArgs = List.map (tType2TypeSig env) tparams in
        let mspec = mk_mref_mspec_in_typ(methodRef, vtypeSig, genArgs) in
        let callInstr = mk_normal_callvirt mspec in
          instrs@[callInstr]

    | TExp_call(v, [], ((mname, [tparam]) as tmref)) when mname="TyApp" -> 
        let instrs = tExp2Instrs (TExp_val v) in
        let vtype = tTypeOftValue env v in  (* vtype should be like All<\'c. Fun<'c,'c>> *)
        let (retType, mname, _, _, _), tfun_arg = match vtype with
          | TType_concrete ((tcn, [targ], _) as tconc) ->
              must (lookup_method env tconc tmref) (fun () -> (spr "Method ref TyApp not found in %s" (Sugar.text_of_path tcn))), targ
          | _ -> raise Impos in
        let vtypeSig = tType2TypeSig env vtype in (* should be just All<> *)
        let vtypeRef = tref_of_typ vtypeSig in
        let retTypeSig = tType2TypeSig env retType in (* should be object *)
        let cc = Callconv(CC_instance, CC_default) in
        let methodRef = mk_mref(vtypeRef, cc, mname, 1, [], retTypeSig) in
        let genArgs = List.map (tType2TypeSig env) [tparam] in
        let mspec = mk_mref_mspec_in_typ(methodRef, vtypeSig, genArgs) in
        let callInstr = mk_normal_callvirt mspec in
        let retTyp = TType_tapp(tfun_arg, tparam) in
          (*           pr "Inserting a cast to type %s\n which normalizes to %s\n" (PrettyTarget.strType retTyp) (PrettyTarget.strType (TargetChecker.normalize retTyp)); *)
        let retTypSig = tType2TypeSig env (TargetChecker.normalize retTyp) in
          instrs@[callInstr; I_castclass retTypSig]
          
    | TExp_val (TVal_obj (tconc, argts, vals)) ->
        (*| TExp_newobj (tconc, argts, vals) ->     *)
        let tn, targs, vargs = tconc in
        let instrsl = List.map (fun v -> tExp2Instrs (TExp_val v)) (vargs@vals) in
          (match tryfind_extern_classdecl env tn with 
             | Some cdecl when cdecl.extqual=Some Sugar.Nullary ->
                 (match cdecl.extends with
                      Some (TType_name(enclosing::ename, _)) -> 
                        let _::dll::_, propertyName = Util.pfx tn in
                        let assref = my_mk_simple_assref dll in
                        let sr = ScopeRef_assembly assref in
(*                         let _ = printfn "%A" ("Making name ", dll::ename) in *)
                        let tref = mk_tref_in_tref(mk_tref(sr, enclosing), Sugar.text_of_path ename)  in
                          (*                         let tref = tClassName2TypeRef env ename in *)
                        let mn = "get_uniq_"^propertyName in
                        let tsig = mk_boxed_typ tref [] in
                        let argTsigs = [] in
                        let retTsig = tsig in
                        let genArgs = [] in
                        let mspec = mk_static_mspec_in_typ (tsig, mn, argTsigs, retTsig, genArgs) in
                        let callInstr = mk_normal_call mspec in
                          [callInstr]
                    | _ -> raise Impos)
             | _ -> 
                 let cdecl = TargetChecker.lookupCdeclByName env.classDeclEnv tn in
                 let formalTargs = cdecl.ptvars@cdecl.tvars in
                 let tsubst = List.rev (snd (List.fold_left (fun (ix, out) t -> ix+1us, TargetChecker.TvarMap(t, TType_index ix)::out) (0us, []) formalTargs)) in
                 let paramTypes = List.map (fun (_,t) -> t) (cdecl.vvars@cdecl.fields) in
                 let paramTypes = List.map (fun t -> 
                                              tType2TypeSig env (TargetChecker.substType tsubst t)) paramTypes in
                 let tref = tClassName2TypeRef env tn (List.length (first_order_tvars formalTargs)) in 
                 let boxity = AsObject in
                   (*                  let paramTypes =  *)
                   (*                    if List.length argts > 0 then List.map (fun t -> tType2TypeSig env t) argts  *)
                   (*                    else List.map (fun v -> tType2TypeSig env (tTypeOftValue env v)) (vargs@vals) in *)
                 let genArgs = List.map (tType2TypeSig env) targs in 
                 let mspec = mk_ctor_mspec (tref, boxity, paramTypes, genArgs) in
                   (List.concat instrsl)@[mk_normal_newobj mspec])
            
    | TExp_bot -> 
        let tref = mk_tref(ecma_mscorlib_scoref, "System.Exception") in
        let mspec = mk_ctor_mspec(tref, AsObject, [mscorlibrefs.typ_String], []) in
          [I_ldstr ("Incomplete pattern match!"); 
           mk_normal_newobj mspec; 
           I_throw]
        
    | ee -> printfn "%A" ("Unexpected expression not translatable to a basic block", ee); raise Impos (* Other expression forms are translated into blocks *) in

  let rec tExp2Code (jump:ILInstr) : tExp -> (ILCodeLabel * ILCode) = function
    | TExp_let (x, e1, e2) -> 
        let e2label, e2code = tExp2Code jump e2 in
        let instrs = 
          let loc = must (lookup_local env x) (fun () -> "Local not found") in
            [I_stloc loc; I_br e2label] in
        let letlabel = generate_code_label () in
        let xbb = ILBasicBlock({bblockLabel=letlabel; bblockInstrs=Array.of_list instrs}) in
        let jxbb = I_br letlabel in
        let e1label, e1code = tExp2Code jxbb e1 in
        let groupBlock = mk_group_block([e2label; letlabel], [e1code;xbb;e2code]) in
          e1label, groupBlock

    | TExp_cond (e1, e2, e3) -> 
        let e2label, e2code = tExp2Code jump e2 in
        let e3label, e3code = tExp2Code jump e3 in
        let brlabel = generate_code_label () in
        let instrs = [I_brcmp (BI_brtrue, e2label, e3label); I_br e3label] in 
        let brcode = ILBasicBlock({bblockLabel=brlabel;
                                 bblockInstrs=Array.of_list instrs}) in
        let e1label, e1code = tExp2Code (I_br brlabel) e1 in
        let groupBlock = mk_group_block([e2label;e3label;brlabel], [e1code;brcode;e2code;e3code]) in
          e1label, groupBlock

    | TExp_isinst (v, branches, def, tt) -> 
        let ldv, localnamex = match v with 
          | TVal_var x -> 
              (match lookup_local env x with 
                 | Some vlocalId -> I_ldloc vlocalId, x
                 | _ -> (match lookup_param env x with 
                             Some paramId -> I_ldarg paramId, x
                           | None -> raise (Ilgen_error (spr "(in class %s): local variable or parameter %s not found" (env.curclassname) (fst x)))))
          | _ -> raise Impos in
        let defLabel, defCode = tExp2Code jump def in 
        let label, internalLabels, codel = List.fold_right 
          (fun (tcn, ttl, tvl, branch) (defLabel, internalLabels, codel) -> 
             let blabel, branchCode = tExp2Code jump branch in (* TODO: load locals *)
             let cdecl = match lookup_cdecl env tcn with 
               | None -> let msg = 
                   spr "Translation error when translating module %s\n\
                        Class %s not found in environment %s" 
                     env.curmodulename (Sugar.text_of_path tcn) (strEnv env) in 
                   raise (Ilgen_error msg)
               | Some cd -> cd in
             let allfields = cdecl.vvars@cdecl.fields in
               if (List.length tvl  <> List.length allfields) then raise Impos;
               let ttc = TType_concrete(tcn, ttl, List.map (fun v -> TVal_var v) tvl) in
               let v' = TVal_var (fst localnamex, ttc) in
               let stlocs = 
                 (List.fold_left (fun out ((fn,_), loc) -> 
                                    let tgtExp = TExp_ldfld(v', fn) in
                                    let ldfld = tExp2Instrs tgtExp in
                                    let loc = must (lookup_local env loc) (fun () -> "Local not found") in
                                      ldfld@(I_stloc loc::out)) [I_br blabel] (List.zip allfields tvl)) in

               let storeLocsLabel = generate_code_label () in
               let storeCode = ILBasicBlock({bblockLabel=storeLocsLabel; bblockInstrs=Array.of_list stlocs}) in
               let testTypeInstr = 
                 (match tryfind_extern_classdecl env tcn with 
                    | Some cdecl when cdecl.extqual=Some Sugar.Nullary ->
                        (match cdecl.extends with
                             Some (TType_name(enclosing::ename, _)) -> 
                               let _::dll::_, propertyName = Util.pfx tcn in
                               let assref = my_mk_simple_assref dll in
                               let sr = ScopeRef_assembly assref in
                               let tref = mk_tref_in_tref(mk_tref(sr, enclosing), Sugar.text_of_path ename)  in
                               let mn = "Is"^propertyName in
                               let tsig = mk_boxed_typ tref [] in
                               let cc = Callconv(CC_instance, CC_default) in
                               let mspec = mk_nongeneric_mspec_in_typ (tsig, cc, mn, [], mscorlibrefs.typ_Bool) in
                               let callInstr = mk_normal_callvirt mspec in
                                 callInstr
                           | _ -> raise Impos)
                    | _ -> 
                        let testTypeSig = tType2TypeSig env (TType_concrete(tcn, ttl, List.map (fun t -> TVal_var t) tvl)) in
                          I_isinst testTypeSig) in
               let instrs = [ldv; 
                             testTypeInstr;
                             I_brcmp (BI_brtrue, storeLocsLabel, defLabel);
                             I_br defLabel] in
               let testLabel = generate_code_label () in
               let testCode = ILBasicBlock({bblockLabel=testLabel; bblockInstrs=Array.of_list instrs}) in
                 testLabel, storeLocsLabel::blabel::defLabel::internalLabels, testCode::storeCode::branchCode::codel)
          branches (defLabel, [], [defCode]) in
        let groupBlock = mk_group_block(internalLabels, codel) in
          label, groupBlock
            
    | e -> (* these are compiled to straight-line sequences of instructions *)
        let instrs = tExp2Instrs e in
        let instrs = match e with 
          | TExp_bot -> instrs
          | _ -> instrs@[jump] in
        let cl = generate_code_label () in
          cl, ILBasicBlock ({bblockLabel=cl; bblockInstrs=(Array.of_list instrs)}) in
    
  let label, code = tExp2Code I_ret exp in
    code

let tExp2ILMethodBody env e = 
  let rec collect_locals out = function
    | TExp_let (x, e1, e2) -> 
        let out = collect_locals (x::out) e1 in
          collect_locals out e2
    | TExp_isinst (v, branches, defbranch, t) -> (* Q: What's the last argument t for? *)
        let out = List.fold_left (fun out (tc, targs, xargs, e) -> 
                                    collect_locals (xargs@out) e) out branches in
          collect_locals out defbranch
    | TExp_cond (e1, e2, e3) -> 
        List.fold_left (fun out e -> 
                          collect_locals out e) out [e1;e2;e3] 
    | _ -> out in
  let localTvars = collect_locals [] e in
  let env = add_local_map env localTvars in
  let locals = List.map (fun (tid, tType) -> 
                           let tsig = tType2TypeSig env tType in
                             mk_local tsig) localTvars in
  let maxStackDepth = DEFAULT_MAX_STACK in
  let code = tExp2Code env e in
    mk_ilmbody (true, locals, maxStackDepth, code, None)


let tMethodDecl2MethodDef env (retType, mname, targs, args, bodyOpt) = 
  let gpds = List.map tVar2GenericParameterDef targs in
  let env = add_generic_params env gpds in
  let plist = List.map (fun (n, t) -> 
                          let ts = tType2TypeSig env t in
                            mk_param (None, ts)) args in
  let env = add_parameter_map env args in
  let retTypeSig =  (* retType is already normalized; so can't tell that it's a TType_tapp *)
    if mname="TyApp" then mscorlibrefs.typ_Object 
    else tType2TypeSig env retType in
  let body, mk = match bodyOpt with
      None -> MethodBody_abstract, Some (MethodKind_virtual {virtFinal=false;
                                                             virtNewslot=true;
                                                             virtStrict=false;
                                                             virtAbstract=true})
    | Some b ->  MethodBody_il (tExp2ILMethodBody env b), None in
  let res = match gpds with
      [] -> 
        mk_nongeneric_virtual_mdef(mname, MemAccess_public, plist, mk_return retTypeSig, body)
    | _ -> 
        mk_generic_virtual_mdef (mname, MemAccess_public, 
                                 gpds, plist, 
                                 mk_return retTypeSig, 
                                 body) in
    match mk with
        None -> res
      | Some mk -> {res with mdKind=mk} 
          
let genericParameterDef2GenericArg env gpd = 
  let tvid = must (lookup_tvar_id env (gpd.gpName, ())) (fun () -> "Tyvar not found") in
    Type_tyvar tvid

let makeConstructorMethodDef env superTsig supers fds = 
  let params, setFieldInstrsL, _ = 
    List.fold_left (fun (params, instrsL, count) (fd:ILFieldDef) ->
                      let param = mk_named_param(fd.Name, fd.Type) in
                      let isStatic, fspec = must (lookup_fieldspec env (fd.Name, ())) (fun () -> "Field spec not found") in
                      if isStatic then raise Impos;
                      let stinstrs = [I_ldarg 0us;
                                      I_ldarg count;
                                      mk_normal_stfld fspec] in
                        param::params, stinstrs::instrsL, count + 1us) ([], [], 1us) fds in
  let rec lookup_super_field ix fds fn = match fds with
    | [] -> lookup_failed "Super argument not found" fn
    | fd::tl when fd.fdName=fn -> ix, fd.fdType
    | _::tl -> lookup_super_field (ix+1us) tl fn in
  let superArgs = List.map (lookup_super_field 1us fds) supers in
  let superCtorSpec = mk_ctor_mspec_for_typ(superTsig, List.map snd superArgs) in
  let ldargs = List.map (fun (ix, _) -> I_ldarg ix) superArgs in
  let superCall = [ldarg_0]@ldargs@[mk_normal_call superCtorSpec] in
  let params = List.rev params in
  let setFields = List.concat (List.rev setFieldInstrsL) in
  let ctorInstrs = superCall@setFields@[I_ret] in
  let code = nonbranching_instrs_to_code ctorInstrs in
  let mb = MethodBody_il (mk_ilmbody (true, [], DEFAULT_MAX_STACK, code, None)) in
    mk_ctor(MemAccess_public, params, mb)

let mk_entry_point env e =
  trace -1977 "Entry point target expression" e;
  let mbody = MethodBody_il (tExp2ILMethodBody {env with eraseUnit=true} e) in
  let mdef = mk_static_nongeneric_mdef ("__entrypoint", 
                                        MemAccess_public, 
                                        [],
                                        mk_return Type_void, 
                                        mbody) in
  let mdef = {mdef with mdEntrypoint=true} in
    mdef

let add_initializers env td tsfds fspecs =
  let add_initializer (fn, t, init) fspec td =  
    match init with
        None -> td
      | Some init -> 
          let code = tExp2Code env init in
          let instrs = match code with 
            | ILBasicBlock bb -> Array.to_list bb.bblockInstrs 
            | _ -> raise (NYI "Static field initializer is non-primitive") in
          let instrs, ret = Util.pfx instrs in
          let stfld = mk_normal_stsfld fspec in
          let instrs = instrs@[I_castclass (fspec.FormalType); stfld] in
            prepend_instrs_to_cctor instrs None td in
    List.fold_right2 add_initializer tsfds fspecs td 
      

let tClassDecl2TypeDef env (tcd:tClassDecl) = 
  let ns, className = tClassName2NamespaceAndString tcd.name in
  let env = {env with curclassname=className} in
  let tda = tVisibility2TypeDefAccess tcd.visibility in
    (*   let _ = match tcd.pvvars with *)
    (*     | [] -> () *)
    (*     | _ -> raise (NYI "Classes with phantom value parameters") in *)
  let fo_tvars = first_order_tvars (tcd.ptvars @ tcd.tvars) in
  let gpds = List.map tVar2GenericParameterDef fo_tvars in
  let env = add_generic_params env gpds in
  let is_phantom n = List.exists (fun (m,_) -> n=m) tcd.pvvars in
  let ext, superArgs = match tcd.extends with 
    | None -> mscorlibrefs.typ_Object, []
    | Some (TType_affine(TType_concrete(cn, tpl, vpl) as t_ext))  
    | Some ((TType_concrete(cn, tpl, vpl) as t_ext))  -> 
        let t_extSig = tType2TypeSig env t_ext in
        let vpl = List.filter (function TVal_var(n,t) -> not (is_phantom n)
                                 | _ -> true) vpl in
        let valParams = List.map (function 
                                    | TVal_var (n, t) -> n 
                                    | _ -> raise (NYI "Non-variable super-class parameters")) vpl in (* field names *)
          t_extSig, valParams

    | Some (TType_name(cn, _)) -> 
        let t_extSig = tType2TypeSig env (TType_concrete(cn, [], [])) in
          t_extSig, []

    | _ -> printfn "%A" tcd.extends; raise Impos in

  let vvars = tcd.vvars in
  let vvfds = List.map (tFieldDecl2FieldDef Instance env) vvars in 
(*   let fields = removeDups (fun (fn1,_) (fn2,_) -> fn1=fn2) tcd.fields in *)
  let fields = tcd.fields in
  let ifds = List.map (tFieldDecl2FieldDef Instance env) fields in
    (* TODO: handle initializer *)
(*   let sfds = removeDups (fun (fn1,_,_) (fn2,_,_) -> fn1=fn2) tcd.staticFields in *)
  let sfds = tcd.staticFields in
  let sfds = List.map (fun (fn, typ, init) -> 
(*                          Printf.printf "Adding static field named %s\n" fn; *)
                         tFieldDecl2FieldDef Static env (fn, typ)) sfds in
  let fds = IL.mk_fdefs (vvfds@ifds@sfds) in
  let ngenerics = List.length gpds in
  let className = if ngenerics <> 0 then spr "%s`%d" className ngenerics else className in
(*     printfn "Generating decl for class %s\n" className; *)
  let thisTypeRef = 
    if className = env.curmodulename then 
      ILTypeRef.Create (ScopeRef_local, [], className) 
    else   
      ILTypeRef.Create (ScopeRef_local, [env.curmodulename], className) in
  let genericArgs = List.map (genericParameterDef2GenericArg env) gpds in
  let thisTypeSig = mk_boxed_typ thisTypeRef genericArgs in
  let fieldSpecs = List.map 
    (fun fd -> 
       let fref =  mk_fref_to_fdef (thisTypeRef, fd) in
       let fspec = mk_fspec (fref, thisTypeSig) in
         fd.fdName, (fd.IsStatic, fspec)) (dest_fdefs fds) in
  let env = add_fieldspec_map env fieldSpecs in
  let mds = List.map (tMethodDecl2MethodDef env) tcd.methods in
  let isabs = (Target.is_abstract tcd.name) || (List.exists (fun md -> match md.mdKind with 
                                                               | MethodKind_virtual k -> k.virtAbstract 
                                                               | _ -> false) mds) in
  let mds = 
    let ctor = makeConstructorMethodDef env ext superArgs (vvfds@ifds) in 
      ctor::mds in
  let mds = 
    if (Sugar.text_of_path tcd.name) = env.curmodulename then 
      match env.entry with
          None -> mds
        | Some e -> 
            let emdef = mk_entry_point env e in 
              emdef::mds 
    else mds in
  let mds = IL.mk_mdefs mds in
(*     printfn "Generating decl for class %s\n" className; *)
  let td = IL.mk_generic_class (className, tda, gpds, ext, [], mds, fds, mk_tdefs [], props, events, atts, TypeInit_beforefield) in
  let is_nested = className <> env.curmodulename in
  let td = match td.tdAccess with 
      TypeAccess_public when is_nested -> 
        {td with tdAbstract=isabs; tdAccess=IL.Nested MemAccess_public}
    | _ -> {td with tdAbstract=isabs} in
    add_initializers env td tcd.staticFields 
      (List.map (fun sfd -> 
                   let fref =  mk_fref_to_fdef (thisTypeRef, sfd) in
                     mk_fspec (fref, thisTypeSig)) sfds)
      
let remove_dups (decls:list<tClassDecl>) = 
  let rec aux (out:list<string * tClassDecl>) (x:list<tClassDecl>) = match x with
    | [] -> List.rev (List.map snd out)
    | hd::tl -> let thisclassName = String.concat "." hd.name in
        match List.tryFind (fun (cn, _) -> cn=thisclassName) out with
            Some _ -> aux out tl
          | None -> aux ((thisclassName, hd)::out) tl in    
    aux [] decls 
      
let addTopLevelFieldDecls (m:tModule) env : env = 
  let top = m.modCdecl in
  let topTref = ILTypeRef.Create (ScopeRef_local, [], Sugar.text_of_path top.name) in
  let topTsig = mk_nongeneric_boxed_typ topTref in
    {env with topfields=Some(topTref, topTsig, top.staticFields)}

let predefined_classes =
  match Target.predefined_classes with 
      dSS::tl -> tl(*  let dSS = {dSS with methods=[]} in dSS::tl *)
                                                      
let tmod_to_absil env (m:tModule) : ILModuleDef  = 
(*   pr "\n**********\n"; *)
(*   pr "tmod_to_absil called with environment %s\n" (strEnv env); *)
(*   pr "\n**********\n"; *)
  let env = {env with curmodulename=String.concat "." m.name} in
  let mname = Sugar.text_of_path m.name in
  let decls = 
    if m.name <> ["Prims"] then 
      Hashtbl.fold (fun n (cd:tClassDecl) (cls:list<tClassDecl>) -> if (is_primitive (List.hd cd.name)) then cls else cls@[cd]) m.decls [] 
        (*List.filter (fun tcd -> not (is_primitive (List.hd tcd.name))) m.decls *)
    else      (*List.filter (fun tcd -> tcd.name <> ["Prims";"unit"]) m.decls *)
      let d = Hashtbl.fold (fun name (cd:tClassDecl) (cls:list<tClassDecl>) -> if (name = "Prims.unit") then cls else cls@[cd]) m.decls [] in
        predefined_classes@d
  in
  let decls = remove_dups decls in
  let env = addTopLevelFieldDecls m env in
(*   let tdefs = IL.mk_tdefs (List.map (tClassDecl2TypeDef env) (m.modCdecl::decls)) in *)
(*     mk_simple_mainmod mname ("Fine-Module-"^mname) true tdefs  *)
  let tdefs = IL.mk_tdefs (List.map (tClassDecl2TypeDef env) decls) in
  let mainclass = match m.extends with
    | Some [mn] -> 
(*         Printf.printf "Module %s extends module %s\n" env.curmodulename mn; *)
        {m.modCdecl with extends=Some(TType_concrete ([mn;mn], [], []))} 
    | _ -> m.modCdecl in
  let mainclass = {(tClassDecl2TypeDef env mainclass) with tdNested=tdefs} in
  let metadataVersion = "2.0.50727.0" in
  let exportedTypes = (mk_exported_types [])  in
  let flags = 0x0 in
    mk_simple_mainmod mname ("Fine-Module-"^mname) true (mk_tdefs [mainclass]) None None flags exportedTypes metadataVersion
      
let outputDir : option<string> ref = ref None      
let write_il (tmod:tModule) (m:ILModuleDef) : unit = 
  let ext, m = match tmod.entry with None -> 
    ".dll", m | Some _ -> ".exe", {m with modulDLL=false} in
  let fname = match m.modulManifest with
    | Some n -> n.manifestName ^ ext 
    | None -> m.modulName ^ ext in 
  let fname = match !outputDir with
      None -> fname
    | Some x -> x ^ "/" ^ fname in
  let _ = Printf.printf "Writing to file %s\n" fname in
  let defaults =  { BinaryWriter.mscorlib=ecma_mscorlib_scoref ;
                    BinaryWriter.pdbfile=None;
                    BinaryWriter.signer=None;
                    BinaryWriter.fixupOverlappingSequencePoints=false } in
    BinaryWriter.WriteILBinary fname defaults m
      
