// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Solver

open Absyn
open Util
open Profiling 
open Tcenv

let ctrs = ref []
let count env = match !ctrs with
  | (mname, c)::tl when (current_module env)=mname -> ctrs := (mname, c+1)::tl
  | l -> ctrs := ((current_module env, 1)::l)
let print_ctrs () =
  List.iter (fun (mname, c) -> if !Util.silent then () else Printf.printf "Proof obls for module %s = %d\n" (Pretty.str_of_lident mname) c) !ctrs
    
let already_disposed = ref false
let dispose_proof_state () = if !already_disposed then () else (Proof.finish(); already_disposed := true ; print_ctrs () )

(* let solver_ctr = new_counter "Solver" *)

let DISABLE_SOLVER = ref false (* HACK! *)

let dump_match_assumptions env = 
  Tcenv.fold_env env (fun _ -> function
                          Tcenv.Binding_match(e1,e2) -> Printf.printf "(%s = %s)\n" (Pretty.strExp e1) (Pretty.strExp e2)
                        | _ -> ()) ()
  
let query (env:Tcenv.env) (formula:typ) : bool  = 
  if !DISABLE_SOLVER then
    (Printf.printf "WARNING!!: When typing %s, admitted proof of refinement %s in a context with\n"
       (Sugar.text_of_lid (Tcenv.current_unit env)) (Pretty.strTyp formula);
     true)
  else if (!extract_proofs && in_tc_phase env) then true 
  else if (not !extract_proofs) && in_tc_phase env then
    (count env;
     let result, popt = Proof.prove_and_extract env formula in result)
  else false
        
let query_equiv (env:Tcenv.env) (e1:exp) (e2:exp) : bool =
  if !DISABLE_SOLVER then
    (Printf.printf "WARNING!!: When typing %s, admitted proof of term equivalence %s = %s in a context with\n" 
       (Sugar.text_of_lid (Tcenv.current_unit env)) (Pretty.strExp e1) (Pretty.strExp e2);
     true)
  else
    ((if (in_derefinement_phase env) then 
        Printf.printf "Warning! Using solver to prove equivalence");
     count env;
     let result, popt = Proof.prove_eq_and_extract env e1 e2 in result)
      
let discharge_proof (env:Tcenv.env) (t:typ) : option<exp> = 
  if not !Util.extract_proofs then Some (withsort Exp_bot t)
  else
    (count env;
     let result, popt = Proof.prove_and_extract env t in 
       if result then bind_opt popt (fun p -> Some (withsort p t))
       else None)
      (*   raise (NYI "Solver.discharge_proof") *)
