// (c) Microsoft Corporation. All rights reserved

module Microsoft.Fine.Tcenv

open Absyn 

type binding =
  | Binding_var of ident * typ
  | Binding_typ of ident * kind
  | Binding_match of exp * exp

(* type env *)
type sigtable = HashMultiMap<string, sigelt>
type phase = | TC | DEREFINE 
type env = {
  phase:phase;
  curmodule: lident;             (* Name of this module---the "principal" subscript *)
  gamma:list<binding>;           (* Main typing environment *)
  externs: signature;            (* Extern declarations accumulated from other modules *)
  e_signature: signature;        (* Type and data constructors for this module *)
  curunit: lident;               (* the name of the current top-level function unit *)
  in_typ_decl: bool;             (* are we currently checking a type declaration*)
  modules:list<lident * modul>;  (* already fully type checked modules *)
  expected_typ:option<typ>;
  recall_expected_typ:option<typ>;
  expected_kind:option<kind>;
  expect_tyapps:bool;
  current_value:option<exp>;
  use_solver:bool;
  kind_level:bool;
  sigtab:sigtable;
  extern_typ:bool;
  expand_externs:bool;
  uvars:list<uvar>
}

val in_tc_phase: env -> bool
val set_tc_phase: env -> env
val in_derefinement_phase: env -> bool
val set_derefinement_phase: env -> env

val initial_env  : lident -> env
val empty_env: env
val modules : env -> list<lident*modul>
val set_solver_flag : env -> bool -> env
val use_solver : env -> bool
val set_kind_level : env -> env
val is_kind_level : env -> bool

exception Not_found_binding of env * Util.Disj<typ',exp'>

val current_module : env -> lident
val set_current_module : env -> lident -> env
val current_unit : env -> lident
val set_current_unit : env -> lident -> env
val in_typ_decl : env -> bool
val set_in_typ_decl : env -> env
val clear_in_typ_decl : env -> env

val lookup_val_decl: env -> lident -> typ
val lookup_bvar : env -> bvar<typ> -> typ
val lookup_lid : env -> lident -> typ
val lookup_fvar : env -> var<typ> -> option<typ>
val lookup_fvar_with_params : env -> var<typ> -> option<(typ * list<tparam>)>
val lookup_operator: env -> ident -> option<typ>

val lookup_btvar : env -> bvar<kind> -> kind
val lookup_typ_const : env -> var<kind> -> kind
val lookup_field_name : env -> lident -> (list<tparam> * typ)
val field_name_exists : env -> lident -> bool
val lookup_record_typ : env -> lident -> (lident * typ * list<tparam> * typ) (* first type is the abbreviation as type name *)
val lookup_record_typ_by_name : env -> lident -> (lident * typ * list<tparam> * typ) 
val get_record_fields: env -> typ -> list<fieldname*typ>
val is_datacon: env -> var<typ> -> bool
val is_prop: env -> typ -> bool

val push_sigelt : env -> sigelt -> env
val push_local_binding : env -> binding -> env
val push_module : env -> modul -> env
val set_expected_typ : env -> typ -> env 
val expected_typ : env -> option<typ> 
val clear_expected_typ : env -> (env * option<typ>)
val recall_expected_typ : env -> option<typ>
val set_expected_kind : env -> kind -> env 
val expected_kind : env -> option<kind> 
val clear_expected_kind : env -> (env * option<kind>)
val uvars_in_env: env -> list<uvar>

val clear_current_state : env -> env

val fold_env : env -> ('a -> binding -> 'a) -> 'a -> 'a
val find: env -> (binding -> bool) -> option<binding>

val set_expect_tyapps : env -> env
val clear_expect_tyapps : env -> env
val expect_tyapps : env -> bool
val dump: env -> unit

val freevars_not_in_env : env -> typ -> (list<bvar<kind>> * list<bvar<typ>>)

val set_current_value : env -> exp -> env
val current_value : env -> option<exp>
val clear_current_value : env -> env

val expand_typ: env -> typ -> typ
val expand_typ_rec: env -> typ -> typ (* expensive! use with care *)
val expand_exp_rec: env -> exp -> exp (* expensive! use with care *)

val tycon_kind_from_tparams : list<tparam> -> kind -> kind
(* deprecated *)
val expand_typ_fix: env -> typ -> typ
val get_sigtab_keys: env -> list<string>

val set_extern_typ : env -> env
val extern_typ : env -> bool
val clear_extern_typ : env -> env
val clear_expand_externs : env -> env
val set_expand_externs : env -> env
